/*******************************************************************************
    Copyright (c) 2016 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

#ifndef _CTRL_NVSWITCH_H_
#define _CTRL_NVSWITCH_H_

#ifdef __cplusplus
extern "C"
{
#endif

// TODO consider making the ctrl interface a module standard

// TODO Document once this is fleshed out more

// TODO consider only making top level ctrl interface visible, hide
//      endpoint implementation from users

typedef NvU32   TBDType;                  //!! Fix these based on IAS and FD

typedef enum _CONNECTION_TYPE
{
    CONNECT_ACCESS_GPU = 0,
    CONNECT_ACCESS_CPU,
    CONNECT_TRUNK_SWITCH,
    CONNECT_ACCESS_SWITCH
} CONNECTION_TYPE;

/*
 * CTRL_NVSWITCH_SET_SWITCH_GLOBAL_CONFIG
 */
typedef struct
{
    TBDType TBDValue0;
    TBDType TBDValue1;
} NVSWITCH_SET_SWITCH_GLOBAL_CONFIG;

/*
 * CTRL_NVSWITCH_SET_SWITCH_PORT_CONFIG
 */
typedef struct tagNVSWITCH_SET_SWITCH_PORT_CONFIG
{
    NvU32           portNum;                    // Physical NPORT
    CONNECTION_TYPE type;
    NvU32           requesterLinkID;              // only applies to endpoints
} NVSWITCH_SET_SWITCH_PORT_CONFIG;

/*
 * CTRL_NVSWITCH_SET_INGRESS_REQUEST_TABLE
 */
typedef struct tagNVSWITCH_INGRESS_REQUEST_ENTRY
{
    NvU32   vcModeValid7_0;
    NvU32   vcModeValid15_8;
    NvU32   vcModeValid17_16;
    NvU32   mappedAddress;
    NvU32   routePolicy;
    NvBool  entryValid;
} NVSWITCH_INGRESS_REQUEST_ENTRY;

typedef struct tagNVSWITCH_SET_INGRESS_REQUEST_TABLE
{
    NvU32   portNum;                    // Physical NPORT
    NvU32   firstIndex;
    NvU32   numEntries;
    NVSWITCH_INGRESS_REQUEST_ENTRY  *entries NV_ALIGN_BYTES(8);
} NVSWITCH_SET_INGRESS_REQUEST_TABLE;

/*
 * CTRL_NVSWITCH_SET_INGRESS_REQUEST_VALID
 */
typedef struct
{
    NvU32   portNum;                    // Physical NPORT
    NvU32   firstIndex;
    NvU32   numEntries;
    NvBool  *entryValid NV_ALIGN_BYTES(8);
} NVSWITCH_SET_INGRESS_REQUEST_VALID;

/*
 * CTRL_NVSWITCH_SET_INGRESS_RESPONSE_TABLE
 */
typedef struct tagNVSWITCH_INGRESS_RESPONSE_ENTRY
{
    NvU32   vcModeValid7_0;
    NvU32   vcModeValid15_8;
    NvU32   vcModeValid17_16;
    NvU32   routePolicy;
    NvBool  entryValid;
} NVSWITCH_INGRESS_RESPONSE_ENTRY;

typedef struct tagNVSWITCH_SET_INGRESS_RESPONSE_TABLE
{
    NvU32   portNum;                    // Physical NPORT
    NvU32   firstIndex;
    NvU32   numEntries;
    NVSWITCH_INGRESS_RESPONSE_ENTRY *entries NV_ALIGN_BYTES(8);
} NVSWITCH_SET_INGRESS_RESPONSE_TABLE;

/*
 * CTRL_NVSWITCH_SET_INGRESS_ATSD_CAM
 */
typedef struct
{
    NvU32   routePolicy;
    NvBool  entryValid;
} NVSWITCH_INGRESS_ATSD_CAM_ENTRY;

typedef struct
{
    NvU32   portNum;                    // Physical NPORT
    NvU32   firstIndex;
    NVSWITCH_INGRESS_ATSD_CAM_ENTRY entries[4];
} NVSWITCH_SET_INGRESS_ATSD_CAM;

/*
 * CTRL_NVSWITCH_INIT_SWITCH_LINK_TRAINING
 */
typedef struct
{
    NvU32   portNum;                    // Physical NPORT
    NvU32   phase;
} NVSWITCH_INIT_SWITCH_LINK_TRAINING;

/*
 * CTRL_NVSWITCH_QUERY_SWITCH_LINK_TRAINING
 */
typedef struct
{
    NvU32   portNum;                    // Physical NPORT
    NvU32   phase;
} NVSWITCH_QUERY_SWITCH_LINK_TRAINING;

/*
 * CTRL_NVSWITCH_INIT_PEER_LINK_TRAINING
 */
typedef struct
{
    CONNECTION_TYPE type;
    NvU32   peerID;
    NvU32   portNum;                    // Physical NPORT
    NvU32   phase;
} NVSWITCH_INIT_PEER_LINK_TRAINING;

/*
 * CTRL_NVSWITCH_QUERY_PEER_LINK_TRAINING
 */
typedef struct
{
    CONNECTION_TYPE type;
    NvU32   peerID;
    NvU32   portNum;                    // Physical NPORT
} NVSWITCH_QUERY_PEER_LINK_TRAINING;

/*
 * CTRL_NVSWITCH_GET_INTERNAL_LATENCY
 */
#define NVSWITCH_MAX_PORTS              18
#define NVSWITCH_EGRESS_HISTOGRAM_SIZE  4

typedef struct
{
    NvU32   egressHistogram[NVSWITCH_MAX_PORTS][NVSWITCH_EGRESS_HISTOGRAM_SIZE];
} NVSWITCH_GET_INTERNAL_LATENCY;

/*
 * CTRL_NVSWITCH_SET_LATENCY_BINS
 */
typedef struct
{
    NvU32   lowThreshold;
    NvU32   medThreshold;
    NvU32   hiThreshold;
} NVSWITCH_LATENCY_BIN;

typedef struct
{
    NVSWITCH_LATENCY_BIN   bin[NVSWITCH_MAX_PORTS];
} NVSWITCH_SET_LATENCY_BINS;

/*
 * CTRL_NVSWITCH_GET_NVLIPT_COUNTERS
 */
typedef struct
{
    NvU64   txCounter0;
    NvU64   txCounter1;
    NvU64   rxCounter0;
    NvU64   rxCounter1;
} NVSWITCH_NVLIPT_COUNTER;

typedef struct
{
    NVSWITCH_NVLIPT_COUNTER   liptCounter[NVSWITCH_MAX_PORTS];
} NVSWITCH_GET_NVLIPT_COUNTERS;

/*
 * CTRL_NVSWITCH_SET_NVLIPT_DEFAULT_COUNTERS
 */
typedef struct
{
    TBDType     TBDValue;
} NVSWITCH_SET_NVLIPT_DEFAULT_COUNTERS;

/*
 * CTRL_NVSWITCH_SET_NVLINK_PERF_MUX
 */
typedef TBDType     NVSWITCH_PERFMON_SELECTOR;

typedef struct
{
    CONNECTION_TYPE type;
    NvU32 registerCount;
    NvU32 registerValue[];
} NVSWITCH_SET_NVLINK_PERF_MUX;

/*
 * CTRL_NVSWITCH_GET_NVLINK_PERF_DATA
 */
#define NVSWITCH_PERFMON_COUNTERS_SIZE  4

typedef struct
{
    NvU32   perfCounters[NVSWITCH_MAX_PORTS][NVSWITCH_PERFMON_COUNTERS_SIZE];
} NVSWITCH_GET_NVLINK_PERF_DATA;

/*
 * CTRL_NVSWITCH_SET_PEER_PERF_MUX
 */
typedef TBDType     NVSWITCH_PEER_PERFMON_SELECTOR;

typedef struct
{
    CONNECTION_TYPE type;
    NvU32 registerCount;
    NvU32 registerValue[];
} NVSWITCH_SET_PEER_PERF_MUX;

/*
 * CTRL_NVSWITCH_GET_PEER_PERF_DATA
 */
#define NVSWITCH_PEER_PERFMON_COUNTERS_SIZE  4

typedef struct
{
    NvU32   perfCounters[NVSWITCH_MAX_PORTS][NVSWITCH_PEER_PERFMON_COUNTERS_SIZE];
} NVSWITCH_GET_NVLINK_PEER_PERF_DATA;

/*
 * CTRL_NVSWITCH_SET_ERROR_NOTIFIER
 */
typedef struct
{
    NvU32   pid;
    NvU32   signal;
} NVSWITCH_SET_ERROR_NOTIFIER;

/*
 * CTRL_NVSWITCH_SET_ERROR_POLICY
 */
#define NVSWITCH_ERROR_CLASSES_SIZE 64

typedef NvU32       NVSWITCH_ERROR_POLICY;

typedef struct
{
    CONNECTION_TYPE type;
    NVSWITCH_ERROR_POLICY   policy[NVSWITCH_ERROR_CLASSES_SIZE];
} NVSWITCH_SET_ERROR_POLICY;

/*
 * CTRL_NVSWITCH_GET_NVLINK_ERROR_DATA
 */
#define NVSWITCH_ERROR_COUNT_SIZE   64

typedef NvU32   NVSWITCH_NVLINK_ARCH_ERROR;
typedef NvU32   NVSWITCH_NVLINK_HW_ERROR;

typedef struct
{
    NvU32   linkNumber;
    NvU32   laneNumber;
    NVSWITCH_NVLINK_ARCH_ERROR  enumArchError;
    NVSWITCH_NVLINK_HW_ERROR    enumHWError;
} NVSWITCH_NVLINK_ERROR;

typedef struct
{
    NvU32                   errorCount;         /* in/out */
    NVSWITCH_NVLINK_ERROR   error[NVSWITCH_ERROR_COUNT_SIZE];
} NVSWITCH_GET_NVLINK_ERROR_DATA;

/*
 * CTRL_NVSWITCH_GET_SWITCH_ERROR_DATA
 */
//typedef NVSWITCH_NVLINK_ARCH_ERROR  NVSWITCH_SWITCH_ARCH_ERROR;
//typedef NVSWITCH_NVLINK_HW_ERROR    NVSWITCH_SWITCH_HW_ERROR;
typedef NVSWITCH_NVLINK_ERROR       NVSWITCH_SWITCH_ERROR;

typedef struct
{
    NvU32                   errorCount;         /* in/out */
    NVSWITCH_SWITCH_ERROR   error[NVSWITCH_ERROR_COUNT_SIZE];
} NVSWITCH_GET_SWITCH_ERROR_DATA;

/*
 * CTRL_NVSWITCH_GET_PEER_ERROR_DATA
 */
//typedef NVSWITCH_NVLINK_ARCH_ERROR  NVSWITCH_SWITCH_ARCH_ERROR;
//typedef NVSWITCH_NVLINK_HW_ERROR    NVSWITCH_SWITCH_HW_ERROR;
typedef NVSWITCH_NVLINK_ERROR       NVSWITCH_PEER_ERROR;

typedef struct
{
    NvU32               errorCount;         /* in/out */
    NVSWITCH_PEER_ERROR error[NVSWITCH_ERROR_COUNT_SIZE];
} NVSWITCH_GET_PEER_ERROR_DATA;

/*
 * CTRL_NVSWITCH_GET_DEVICE_INFO_PARAMS
 *
 * This command retrieves device information a nvswitch device.
 *
 *   instance [in]
 *     Which device instance to return information for
 *
 *   linkMask
 *     Mask of active links for this device
 *   pciInfo
 *     PCI bus information for this device
 *
 * An error will be returned if selected device was not found.
 */
typedef struct 
{
    NvU32 instance;

    NvU32 linkMask;
    struct nvlink_pci_info pciInfo;
} CTRL_NVSWITCH_GET_DEVICE_INFO_PARAMS;

/*
 * CTRL_NVSWITCH_GET_OS_HANDLE
 *
 * This command retrieves the OS handle for the given info.  It can only
 * be called for internal kernel driver components like MODS.
 *
 *   instance
 *     Which device instance to return information for
 *
 *     An error will be returned if selected device was not found.
 */
typedef struct
{
    NvU32 instance;
    void *os_handle NV_ALIGN_BYTES(8);
} CTRL_NVSWITCH_GET_OS_HANDLE_PARAMS;

/*
 * MODS and common nvlink control call interface
 */
#define CTRL_NVSWITCH_GET_OS_HANDLE                     0x00
#define CTRL_NVSWITCH_GET_DEVICE_INFO                   0x01
#define CTRL_NVLINK_NVSWITCH_GET_NVLINK_CAPS            0x02
#define CTRL_NVLINK_NVSWITCH_GET_NVLINK_STATUS          0x03
#define CTRL_NVLINK_NVSWITCH_GET_COUNTERS               0x04
#define CTRL_NVLINK_NVSWITCH_CLEAR_COUNTERS             0x05
#define CTRL_NVLINK_NVSWITCH_GET_ERR_INFO               0x06
#define CTRL_NVLINK_NVSWITCH_INJECT_SL_ERR              0x07

/*
 * Switch Kernel Driver/Fabric Manager Interface IOCTLs
 */
#define CTRL_NVSWITCH_SET_SWITCH_GLOBAL_CONFIG          0x20
#define CTRL_NVSWITCH_SET_SWITCH_PORT_CONFIG            0x21
#define CTRL_NVSWITCH_SET_INGRESS_REQUEST_TABLE         0x22
#define CTRL_NVSWITCH_SET_INGRESS_REQUEST_VALID         0x23
#define CTRL_NVSWITCH_SET_INGRESS_RESPONSE_TABLE        0x24
#define CTRL_NVSWITCH_SET_INGRESS_ATSD_CAM              0x25

// NvSwitch Link Training
#define CTRL_NVSWITCH_INIT_SWITCH_LINK_TRAINING         0x30
#define CTRL_NVSWITCH_QUERY_SWITCH_LINK_TRAINING        0x31
#define CTRL_NVSWITCH_INIT_PEER_LINK_TRAINING           0x32
#define CTRL_NVSWITCH_QUERY_PEER_LINK_TRAINING          0x33

// NvSwitch Performance Metric Control and Collection
#define CTRL_NVSWITCH_GET_INTERNAL_LATENCY              0x40

#define CTRL_NVSWITCH_SET_LATENCY_BINS                  0x41
#define CTRL_NVSWITCH_GET_NVLIPT_COUNTERS               0x42
#define CTRL_NVSWITCH_SET_NVLIPT_DEFAULT_COUNTERS       0x43

#define CTRL_NVSWITCH_SET_SWITCH_PERFMON_MUX            0x44
#define CTRL_NVSWITCH_GET_SWITCH_PERFMON_COUNTERS       0x45
#define CTRL_NVSWITCH_SET_PEER_PERFMON_MUX              0x46
#define CTRL_NVSWITCH_GET_PEER_PERFMON_COUNTERS         0x47

// NvSwitch Error Data Control and Collection
#define CTRL_NVSWITCH_SET_ERROR_NOTIFIER                0x50
#define CTRL_NVSWITCH_SET_ERROR_POLICY                  0x51
#define CTRL_NVSWITCH_GET_NVLINK_ERROR_DATA             0x52
#define CTRL_NVSWITCH_GET_SWITCH_ERROR_DATA             0x53
#define CTRL_NVSWITCH_GET_PEER_ERROR_DATA               0x54

#ifdef __cplusplus
}
#endif

#endif //_CTRL_NVSWITCH_H_
